﻿
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;

// To compute the eigenvalues of a random matrix bys the QR transforamtion
// www.mosismath.com


namespace WindowsFormsApplication1
{
    public partial class MainWin : Form
    {
        public const int MaxOrder = 8;
        public int actOrder;
        public double[] cos;
        public double[] sin;
        public double[] xv;
        public double[,] y;
        public double[,] oldVal;
        public double[,] oldEigenVal;
        double phi = 0;
        int count = 0;
        public bool bShift = true;
        TMatrix o;  // origing matrix
        TMatrix m;  // matrix to compute the eigenvalues from
        TMatrix r;  // rotation matrix
        TMatrix eg; // eigenvector matrix

        TGauss gauss;

        public void CalcRMatrixHess(int p, int q)
        {
            int i, j;
            double radius = Math.Sqrt(m.a[p, p - 1].real * m.a[p, p - 1].real + m.a[q, p - 1].real * m.a[q, p - 1].real);
            for (i = 0; i < m.size; i++)
            {
                for (j = 0; j < m.size; j++)
                {
                    r.a[i, j].real = 0.0;
                }
                r.a[i, i].real = 1.0;
            }
            if (radius != 0.0)
            {
                if (m.a[p, p - 1].real == 0.0)
                {
                    r.a[p, p].real = 0.0;
                    r.a[p, q].real = 1.0;
                }
                else
                {
                    //r.a[p, p].real = Math.Abs(m.a[p, p - 1].real) / radius;
                    //r.a[p, q].real = -Math.Sign(m.a[p, p - 1].real) * m.a[q, p - 1].real / radius;
                    r.a[p, p].real = m.a[p, p - 1].real / radius;
                    r.a[p, q].real = - m.a[q, p - 1].real / radius;
                }
                r.a[q, q].real = r.a[p, p].real;
                r.a[q, p].real = -r.a[p, q].real;
            }
        }


        public void CalcRMatrixQR(int p, int q)
        {
            int i, j;
            double radius;
            radius = Math.Sqrt(m.a[p, p].real * m.a[p, p].real + m.a[q, p].real * m.a[q, p].real);
            for (i = 0; i < m.size; i++)
            {
                for (j = 0; j < m.size; j++)
                {
                    r.a[i, j].real = 0.0;
                }
                r.a[i, i].real = 1.0;
            }
            if (radius != 0.0)
            {
                if (m.a[p, p].real == 0.0)
                {
                    cos[p] = 0.0;
                    sin[p] = 1.0;
                }
                else
                {
                    cos[p] = m.a[p, p].real / radius;
                    sin[p] = -m.a[q, p].real / radius;
                }
                r.a[p, p].real = cos[p];
                r.a[p, q].real = sin[p];
                r.a[q, q].real = r.a[p, p].real;
                r.a[q, p].real = -r.a[p, q].real;
            }
        }


        public void CalcRMatrixRQ(int p, int q)
        {
            int i, j;
            for (i = 0; i < m.size; i++)
            {
                for (j = 0; j < m.size; j++)
                {
                    r.a[i, j].real = 0.0;
                }
                r.a[i, i].real = 1.0;
            }
            r.a[p, p].real = cos[p];
            r.a[p, q].real = sin[p];
            r.a[q, q].real = r.a[p, p].real;
            r.a[q, p].real = -r.a[p, q].real;
        }


        public void RunHessenberg(int order)
        {
            int p, q;
            for (p = 1; p < order - 1; p++)
            {
                for (q = p + 1; q < order; q++)
                {
                    if (Math.Abs(m.a[q, p - 1].real) > 1E-200)
                    {
                        CalcRMatrixHess(p, q);
                        m.MultTranspMatrixLeft(r, p, q);
                        m.MultMatrixRight(r, p, q);
                    }
                }
            }
        }


        public void RunQR(int order)
        {
            int i, p, q;
            if (bShift)
                phi = m.a[order - 1, order - 1].real;
            else
                phi = 0.0;
            for (i = 0; i < order; i++)
                m.a[i, i].real = m.a[i, i].real - phi;
            for (p = 0; p < order - 1; p++)
            {
                q = p + 1;
                if (Math.Abs(m.a[q, p].real) > 1E-200)
                {
                    CalcRMatrixQR(p, q);
                    m.MultTranspMatrixLeft(r, p, q);
                }
            }
        }


        public void RunRQ(int maxOrder)
        {
            int i, p, q;
            for (p = 0; p < maxOrder - 1; p++)
            {
                q = p + 1;
                CalcRMatrixRQ(p, q);
                m.MultMatrixRight(r, p, q);
            }
            for (i = 0; i < maxOrder; i++)
                m.a[i, i].real = m.a[i, i].real + phi;
        }


        public void CalcEigenvals22()
        {
            double a, b, c;
            a = 1.0;
            b = -y[0, 0] - y[1, 1];
            c = y[0, 0] * y[1, 1] - y[0, 1] * y[1, 0];

            if ((b * b) >= (4.0 * a * c))
            {
                y[0, 0] = (-b + Math.Sqrt((b * b) - (4.0 * a * c))) / 2.0 / a;
                y[1, 1] = (-b - Math.Sqrt((b * b) - (4.0 * a * c))) / 2.0 / a;
                y[1, 0] = 0.0;
                y[0, 1] = 0.0;
            }
            else
            {
                y[0, 0] = -b / 2.0 / a;
                y[1, 1] = y[0, 0];
                y[0, 1] = Math.Sqrt((4.0 * a * c) - (b * b)) / 2.0 / a;
                y[1, 0] = -y[0, 1];
            }
        }


        public bool IsDeltoOk(int order)
        {
            double delta = 10.0;
            double deltaEigen = 10.0;
            bool result = false;
            if (Math.Abs(m.a[order - 1, order - 2].real) < 1E-20) // real eigenvalue
            {
                delta = m.a[order - 1, order - 1].real - oldVal[0, 0];
                oldVal[0, 0] = m.a[order - 1, order - 1].real;
                result = Math.Abs(delta) < 1E-10;
            }
            else
            {
                // convergence of the values themselves
                y[0, 0] = m.a[order - 2, order - 2].real;
                y[0, 1] = m.a[order - 2, order - 1].real;
                y[1, 0] = m.a[order - 1, order - 2].real;
                y[1, 1] = m.a[order - 1, order - 1].real;

                delta = (y[0, 0] * y[0, 0] + y[0, 1] * y[0, 1]) -
                             (oldVal[0, 0] * oldVal[0, 0] + oldVal[0, 1] * oldVal[0, 1]) +
                             (y[1, 0] * y[1, 0] + y[1, 1] * y[1, 1]) -
                             (oldVal[1, 0] * oldVal[1, 0] + oldVal[1, 1] * oldVal[1, 1]);
                oldVal[0, 0] = y[0, 0];
                oldVal[0, 1] = y[0, 1];
                oldVal[1, 0] = y[1, 0];
                oldVal[1, 1] = y[1, 1];
                // convergence of the eigenvalues
                CalcEigenvals22();

                deltaEigen = (y[0, 0] * y[0, 0] + y[0, 1] * y[0, 1]) -
                             (oldEigenVal[0, 0] * oldEigenVal[0, 0] + oldEigenVal[0, 1] * oldEigenVal[0, 1]) +
                             (y[1, 0] * y[1, 0] + y[1, 1] * y[1, 1]) -
                             (oldEigenVal[1, 0] * oldEigenVal[1, 0] + oldEigenVal[1, 1] * oldEigenVal[1, 1]);
                oldEigenVal[0, 0] = y[0, 0];
                oldEigenVal[0, 1] = y[0, 1];
                oldEigenVal[1, 0] = y[1, 0];
                oldEigenVal[1, 1] = y[1, 1];
                if (Math.Abs(delta) > Math.Abs(deltaEigen))
                {
                    result = Math.Abs(deltaEigen) < 1E-10;
                }
                else
                {
                    result = Math.Abs(delta) < 1E-10;
                }
            }
            return result;
        }


        public MainWin()
        {
            int i;
            o = new TMatrix(MaxOrder); // origin Matrix
            m = new TMatrix(MaxOrder); // Matrix to compute the eigenvalues
            r = new TMatrix(MaxOrder); // rotation Matrix
            eg = new TMatrix(MaxOrder); // Matrix of eigenvectors
            //y = new TMatrix(2);
            y = new double[,] { { 0.0, 0.0 }, { 0.0, 0.0 } };
            oldVal = new double[,] { { 0.0, 0.0 }, { 0.0, 0.0 } };
            oldEigenVal = new double[,] { { 0.0, 0.0 }, { 0.0, 0.0 } };
            cos = new double[MaxOrder];
            sin = new double[MaxOrder];
            xv = new double[MaxOrder];
           // DataGridViewCell cell;

            gauss = new TGauss(MaxOrder, eg);

            InitializeComponent();

            GMatrixA.ColumnCount = MaxOrder;
         //   GMatrixA.RowCount = MaxOrder;
            for (i = 0; i < GMatrixA.ColumnCount; i++)
                GMatrixA.Columns[i].Width = 160;
            GEigenvals.ColumnCount = MaxOrder;
            GEigenvals.RowCount = 2;

            cbShift.Checked = bShift;

            
            // Sample matrix
            GMatrixA.Rows.Add(new object[] { 1 ,  1 ,  3 , -1 , -5 , -2,   1 ,  5 });
            GMatrixA.Rows.Add(new object[] {-4 , -3 ,  2 ,  5 , -3 ,  2 ,  1 , -1 });
            GMatrixA.Rows.Add(new object[] {-4 , -3 , -1 ,  5 , -2  , 0 ,  2 ,  3 });
            GMatrixA.Rows.Add(new object[] {-2  , 3 , -5  , 0 ,  4 ,  0,  -1 ,  5 });
            GMatrixA.Rows.Add(new object[] { 0 ,  4  , 1 , -2 , -1 ,  0,   3,  -1 });
            GMatrixA.Rows.Add(new object[] {-3 ,  0 ,  4 ,  3 ,  0 , -2 ,  4 , -3 });
            GMatrixA.Rows.Add(new object[] { 2 ,  4 , -3 , -2 , -1 ,  1,  -3 ,  0 });
            GMatrixA.Rows.Add(new object[] { 1 ,  2 ,  0 , -1 ,  4 , -2 ,  4 ,  1 });
            //real Eigenvalues:   { -4,845938894957074 ;  -3,2266094228662903 }
            //complex Eigenvalues: { -2,089030322062721-2,400422676707508î ;  -2,089030322062721+2,400422676707508î ;  -1,842477427126424+8,030106840751566î ;  -1,842477427126423-8,030106840751568î ;  3,9677819081008265-0,45220039382670735î ;  3,967781908100827+0,4522003938267094î }

        }


        private void cbShift_CheckedChanged(object sender, EventArgs e)
        {
            bShift = cbShift.Checked;
        }


        private void btHessenberg_MouseClick(object sender, MouseEventArgs e)
        {
            int i, k;
            double[] buf = new double[MaxOrder];
            TMatrix a = new TMatrix(MaxOrder);
            actOrder = MaxOrder;
            DataGridViewCell cell;

            for (i = 0; i < MaxOrder; i++)
            {
                for (k = 0; k < MaxOrder; k++)
                {
                    cell = GMatrixA[i, k];  // get cell to access
                    m.a[k, i].real = Convert.ToDouble(cell.Value);
                    o.a[k, i].real = m.a[k, i].real;
                }
            }
            RunHessenberg(MaxOrder);


            for (i = 0; i < MaxOrder; i++)
            {
                for (k = 0; k < MaxOrder; k++)
                {
                    cell = GMatrixA[i, k];  // get cell to access
                    cell.Value = Convert.ToString(m.a[k, i].real);

                }
            }
        }

        private void btQRdecomp_Click(object sender, EventArgs e)
        {
            int i, k;
            DataGridViewCell cell;
            double[] oldVal = new double[actOrder];
            double delta = 10.0;
            for (i = 0; i < actOrder; i++)
                oldVal[i] = m.a[i, i].real;
            while (actOrder > 1)
            {
                delta = 0;
                count++;
                RunQR(actOrder);
                RunRQ(actOrder);
                for (i = 0; i < actOrder; i++)
                {
                    for (k = 0; k < actOrder; k++)
                    {
                        cell = GMatrixA[i, k];  // get cell to access
                        cell.Value = Convert.ToString(m.a[k, i].real);
                    }
                    delta = delta + Math.Abs(oldVal[i] - m.a[i, i].real);
                    oldVal[i] = m.a[i, i].real;
                }
                if (IsDeltoOk(actOrder))
                {
                    if (actOrder > 1)
                    {
                        if (Math.Abs(m.a[actOrder - 1, actOrder - 2].real) < 1E-20) // real eigenvalue
                        {
                            cell = GEigenvals[actOrder - 1, 0];  // get cell to access
                            cell.Value = Convert.ToString(Math.Round(m.a[actOrder - 1, actOrder - 1].real, 4));
                            cell = GEigenvals[actOrder - 1, 1];  // get cell to access
                            cell.Value = "";
                            actOrder--;
                        }
                        else
                        {
                            if (Math.Round(y[1, 0], 4) == -Math.Round(y[0, 1], 4))
                            {
                                cell = GEigenvals[actOrder - 2, 1];  // get cell to access
                                if (Math.Round(y[1, 0], 4) != 0) // no real eigenvalue
                                    cell.Value = Convert.ToString(Math.Round(y[1, 0], 4));
                                cell = GEigenvals[actOrder - 2, 0];  // get cell to access
                                cell.Value = Convert.ToString(Math.Round(y[0, 0], 4));

                                cell = GEigenvals[actOrder - 1, 0];  // get cell to access
                                cell.Value = Convert.ToString(Math.Round(y[1, 1], 4));
                                cell = GEigenvals[actOrder - 1, 1];  // get cell to access
                                if (Math.Round(y[0, 1], 5) != 0)  // no real eigenvalue
                                    cell.Value = Convert.ToString(Math.Round(y[0, 1], 4));
                            }
                            else
                            {
                                cell = GEigenvals[actOrder - 2, 1];  // get cell to access
                                cell.Value = Convert.ToString(Math.Round(m.a[actOrder - 1, actOrder - 2].real, 4));
                                cell = GEigenvals[actOrder - 2, 0];  // get cell to access
                                cell.Value = Convert.ToString(Math.Round(m.a[actOrder - 2, actOrder - 2].real, 4));

                                cell = GEigenvals[actOrder - 1, 0];  // get cell to access
                                cell.Value = Convert.ToString(Math.Round(m.a[actOrder - 1, actOrder - 1].real, 4));
                                cell = GEigenvals[actOrder - 1, 1];  // get cell to access
                                cell.Value = Convert.ToString(Math.Round(m.a[actOrder - 2, actOrder - 1].real, 4));
                            }
                            actOrder -= 2;
                        }
                    }
                }
            }
            if (actOrder == 1)
            {
                cell = GEigenvals[actOrder - 1, 0];  // get cell to access
                cell.Value = Convert.ToString(Math.Round(m.a[actOrder - 1, actOrder - 1].real, 4));
                cell = GEigenvals[actOrder - 1, 1];  // get cell to access
                cell.Value = "";
            }
            tbCount.Text = count.ToString();
        }

        private void btEigenvect_Click(object sender, EventArgs e)
        {
            int i, k, actCol = 2;
            TComplex EigenVal;
            DataGridViewCell cell;
            for (actCol = 0; actCol < MaxOrder; actCol++)
            {
                cell = GEigenvals[actCol, 0];
                EigenVal.real = Convert.ToDouble(cell.Value);
                double[] buf = new double[MaxOrder];
                cell = GEigenvals[actCol, 1];
                if (cell.Value != null)
                {
                    if (cell.Value.ToString() != "")
                        EigenVal.imag = Convert.ToDouble(cell.Value);
                    else
                        EigenVal.imag = 0;
                }
                else
                    EigenVal.imag = 0;
                for (i = 0; i < MaxOrder; i++)
                {
                    for (k = 0; k < MaxOrder; k++)
                    {
                        eg.a[k, i].real = o.a[k, i].real;
                        eg.a[k, i].imag = o.a[k, i].imag;
                    }
                    eg.a[i, i] = MathComp.Diff(eg.a[i, i], EigenVal);
                    eg.y[i].real = 0.0;
                    eg.y[i].imag = 0.0;
                    eg.x[i].real = 0.0;
                    eg.x[i].imag = 0.0;
                }

                if (gauss.Eliminate())
                {
                    gauss.Solve();
                    for (i = 0; i < MaxOrder; i++)
                    {
                        cell = GMatrixA[actCol, i];
                        if (eg.x[i].imag == 0.0)
                            cell.Value = Convert.ToString(Math.Round(eg.x[i].real, 4));
                        else
                        {
                            if (eg.x[i].imag > 0.0)
                                cell.Value = Convert.ToString(Math.Round(eg.x[i].real, 4)) + " j" + Convert.ToString(Math.Round(eg.x[i].imag, 4));
                            else
                                cell.Value = Convert.ToString(Math.Round(eg.x[i].real, 4)) + " - j" + Convert.ToString(Math.Round(-eg.x[i].imag, 4));
                        }
                    }
                }
                else
                    MessageBox.Show("Calculation error");
            }
        }
    }
}
